#  _  __  
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|            
#
# Keeper Commander 
# Contact: ops@keepersecurity.com
#
import json
import shutil
from typing import Tuple, List, Union, Optional

from colorama import init, Fore, Back, Style
from tabulate import tabulate

from prompt_toolkit.styles.named_colors import NAMED_COLORS

from . import __version__
from .subfolder import BaseFolderNode

init()


class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[33m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'
    HIGHINTENSITYRED = '\033[1;91m'
    WHITE = '\033[0;37m'
    HIGHINTENSITYWHITE = '\033[97m'


def keeper_color_to_prompt(color):   # type: (Optional[str]) -> str
    if color is None:
        return ''
    if color == 'red':
        return 'fg:ansired'
    if color == 'green':
        return 'fg:ansigreen'
    if color == 'blue':
        return 'fg:ansiblue'
    if color == 'orange':
        return 'fg:ansimagenta'
    if color == 'yellow':
        return 'fg:ansiyellow'
    if color == 'gray':
        return 'fg:ansibrightblack'
    return ''

def keeper_colorize(text, color):
    if color == 'red':
        return f'{Fore.RED}{text}{Fore.RESET}'
    if color == 'green':
        return f'{Fore.GREEN}{text}{Fore.RESET}'
    if color == 'blue':
        return f'{Fore.BLUE}{text}{Fore.RESET}'
    if color == 'orange':
        return f'{Fore.MAGENTA}{text}{Fore.RESET}'
    if color == 'yellow':
        return f'{Fore.YELLOW}{text}{Fore.RESET}'
    if color == 'gray':
        return f'{Fore.LIGHTBLACK_EX}{text}{Fore.RESET}'
    return text


def welcome():
    lines = []    # type: List[Union[str, Tuple[str, str]]]

    lines.append( r'         /#############/   /#\ ')
    lines.append( r'        /#############/   /###\      _    __  _______  _______  ______   _______  ______ (R)')
    lines.append( r'       /#############/   /#####\    | |  / / |  ____/ |  ____/ |  ___ \ |  ____/ |  ___ \ ')
    lines.append( r'      /######/           \######\   | | / /  | |____  | |____  | | __| || |____  | | __| | ')
    lines.append( r'     /######/             \######\  | |< <   |  ___/  |  ___/  | |/___/ |  ___/  | |/_  / ')
    lines.append( r'    /######/               \######\ | | \ \  | |_____ | |_____ | |      | |_____ | |  \ \ ')
    lines.append( r'    \######\               /######/ |_|  \_\ |_______||_______||_|      |_______||_|   \_\ ')
    lines.append((r'     \######\             /######/', r'     ____                                          _ '))
    lines.append((r'      \######\           /######/ ', r'   /  ___|___  _ __ ___  _ __ ___   __ _ _ __   __| | ___ _ __ '))
    lines.append((r'       \#############\   \#####/  ', r"  /  /   / _ \| '_ ` _ \| '_ ` _ \ / _` | '_ \ / _` |/ _ \ '__| "))
    lines.append((r'        \#############\   \###/   ', r'  \  \__| (_) | | | | | | | | | | | (_| | | | | (_| |  __/ | '))
    lines.append((r'         \#############\   \#/    ', r'   \_____\___/|_| |_| |_|_| |_| |_|\__,_|_| |_|\__,_|\___|_| '))
    lines.append('')

    try:
        width = shutil.get_terminal_size(fallback=(160, 50)).columns
    except:
        width = 160
    print(Style.RESET_ALL)
    print(Back.BLACK + Style.BRIGHT + '\n')
    for line in lines:
        if isinstance(line, str):
            if len(line) > width:
                line = line[:width]
            print('\033[2K' + Fore.LIGHTYELLOW_EX + line)
        elif isinstance(line, tuple):
            yellow_line = line[0] if len(line) > 0 else ''
            white_line = line[1] if len(line) > 1 else ''
            if len(yellow_line) > width:
                yellow_line = yellow_line[:width]
            if len(yellow_line) + len(white_line) > width:
                if len(yellow_line) < width:
                    white_line = white_line[:width - len(yellow_line)]
                else:
                    white_line = ''
            print('\033[2K' + Fore.LIGHTYELLOW_EX + yellow_line + Fore.LIGHTWHITE_EX + white_line)

    print('\033[2K' + Fore.LIGHTBLACK_EX + f'{("v" + __version__):>93}\n' + Style.RESET_ALL)


def formatted_records(records, **kwargs):
    """Display folders/titles/uids for the supplied shared folders"""
    params = None
    if 'params' in kwargs:
        params = kwargs['params']

    # Sort by title
    records.sort(key=lambda x: x.title.lower(), reverse=False)

    def abbreviate_text(text: str, chars_num: int):
        if 'verbose' in kwargs and kwargs['verbose']:
            return text
        else:
            return text if len(text) < chars_num else text[:chars_num] + '...'

    if len(records) > 0:
        shared_folder = None
        if 'folder' in kwargs and params is not None:
            fuid = kwargs['folder']
            if fuid in params.folder_cache:
                folder = params.folder_cache[fuid]
                if folder.type in {BaseFolderNode.SharedFolderType, BaseFolderNode.SharedFolderFolderType}:
                    if folder.type == BaseFolderNode.SharedFolderFolderType:
                        fuid = folder.shared_folder_uid
                else:
                    fuid = None
                if fuid and fuid in params.shared_folder_cache:
                    shared_folder = params.shared_folder_cache[fuid]

        table = [[i + 1, r.record_uid, abbreviate_text(r.record_type, 32), abbreviate_text(r.title, 32), r.login, abbreviate_text(r.login_url, 32)] for i, r in enumerate(records)]
        headers = ["#", 'Record UID', 'Type', 'Title', 'Login', 'URL']
        if shared_folder and 'records' in shared_folder:
            headers.append('Permissions')
            for row in table:
                permissions = ''
                for sfr in shared_folder['records']:
                    if sfr['record_uid'] == row[1]:
                        if sfr['can_edit']:
                            permissions = 'Can Edit'
                        if sfr['can_share']:
                            if permissions:
                                permissions += ' & Share'
                            else:
                                permissions = 'Can Share'
                        if not permissions:
                            permissions = 'Read Only'
                        break
                row.append(permissions)

        print(tabulate(table, headers=headers))


def formatted_shared_folders(shared_folders, **kwargs):
    """Display folders/titles/uids for the supplied records"""

    # Sort by folder+title
    shared_folders.sort(key=lambda x: (x.name if x.name else ' ').lower(), reverse=False)

    if len(shared_folders) > 0:

        table = [[i + 1, sf.shared_folder_uid, sf.name] for i, sf in enumerate(shared_folders)]
        print(tabulate(table, headers=["#", 'Shared Folder UID', 'Name']))

        print('')

    skip_details = kwargs.get('skip_details') or False
    # Under 5 recs, just display on the screen
    if len(shared_folders) < 5 and not skip_details:
        for sf in shared_folders:
            sf.display()


def formatted_teams(teams, **kwargs):
    """Display names/uids for the supplied teams"""

    # Sort by name
    teams.sort(key=lambda x: (x.name if x.name else ' ').lower(), reverse=False)

    if len(teams) > 0:

        table = [[i + 1, team.team_uid, team.name] for i, team in enumerate(teams)]
        print(tabulate(table, headers=["#", 'Team UID', 'Name']))

        print('')

    skip_details = kwargs.get('skip_details') or False
    # Under 5 recs, just display on the screen
    if len(teams) < 5 and not skip_details:
        for team in teams:
            team.display()


def formatted_history(history):
    """ Show the history of commands"""

    if not history: return
    if len(history) == 0: return

    print('')
    print('Command history:')
    print('----------------')

    for h in history:
        print(h)

    print('')


def print_record(params, record_uid):
    """ Show record content """

    try:
        cached_rec = params.record_cache[record_uid]
    except KeyError as e:
        raise Exception('Record not found: ' + record_uid)
    data = json.loads(cached_rec['data_unencrypted'].decode('utf-8'))
    print(data)



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2023 Keeper Security Inc.
# Contact: commander@keepersecurity.com
#
# This example also pulls configuration from config.json
#
# Usage:
#   1. initialize Keeper virtual environment
#   2. copy "config.json" file next to this script
#   3. run "python thycotic_records.py". Optional "--debug" parameter enables debug output
#   4. run "sqlite3 thycotic_records.db" to match records

"""
-- Get number of records in Keeper
select count(*) from KeeperRecord;

-- Get number of records in Thycotic Secret Server
select count(*) from ThycoticSecret;

-- List records missing in Keeper
select t.* from ThycoticSecret t left outer join KeeperRecord k ON t.title=k.title and t.path=k.path where k.uid IS NULL;

-- List records missing in Thycotic Secret Server
select k.* from KeeperRecord k left outer join ThycoticSecret t ON t.title=k.title and t.path=k.path where t.id IS NULL;

-- List matching records
select t.id, k.* from KeeperRecord k inner join ThycoticSecret t ON t.title=k.title and t.path=k.path;
"""

import argparse
import getpass
import logging
import os
import sqlite3
import sys
from typing import List

from keepercommander.importer.thycotic import thycotic
from keepercommander.__main__ import get_params_from_config
from keepercommander import api, vault, subfolder
from keepercommander.storage import sqlite_dao, sqlite


parser = argparse.ArgumentParser(description='Matches Keeper and Thycotic records')
parser.add_argument('--debug', action='store_true', help='Enables debug logging')
opts, flags = parser.parse_known_args(sys.argv[1:])

logging.basicConfig(level=logging.DEBUG if opts.debug is True else logging.DEBUG, format='%(message)s')

if not os.path.isfile('config.json'):
    logging.error('Please copy config.json file to the current folder.')
    exit(1)

my_params = get_params_from_config('config.json')
logging.info('Connecting to "%s" as "%s"', my_params.server, my_params.user)
api.login(my_params)
if not my_params.session_token:
    exit(1)
api.sync_down(my_params)

logging.info('Connecting to Thycotic Secret Server')
while True:
    thycotic_url = input('{0}: '.format('Thycotic Server URL'.rjust(32)))
    # thycotic_url = 'https://Thycotic/SecretServer'
    thy_auth = thycotic.ThycoticAuth(thycotic_url)

    thycotic_username = input('{0}: '.format('Thycotic Username'.rjust(32)))
    # thycotic_username = 'thycotic'

    thycotic_password = getpass.getpass('{0}: '.format('Thycotic Password'.rjust(32)))
    try:
        thy_auth.authenticate(thycotic_username, thycotic_password)
        break
    except Exception as e:
        logging.warning(e)


thy_folders = thycotic.ThycoticMixin.get_folders(thy_auth, skip_permissions=True)
logging.info(f'Thycotic: Loaded %d folders', len(thy_folders))
thy_secrets = thy_auth.thycotic_search('/v2/secrets')
logging.info(f'Thycotic: Loaded %d secrets', len(thy_secrets))

# root_folders = [x['id'] for x in thy_folders.values() if x['folderName'] == x['folderPath']]
# for folder_id in root_folders:
#     # query = f'/v2/secrets?filter.folderId={folder_id}&filter.includeSubFolders=true'

class KeeperRecord:
    def __init__(self):
        self.uid = ''
        self.title = ''
        self.path = ''

class ThycoticSecret:
    def __init__(self):
        self.id = 0
        self.title = ''
        self.path = ''

keeper_schema = sqlite_dao.TableSchema.load_schema(KeeperRecord, 'uid', indexes={'Record': ['title', 'path']})
thycotic_schema = sqlite_dao.TableSchema.load_schema(ThycoticSecret, 'id', indexes={'Record': ['title', 'path']})

sqlite_file = 'thycotic_records.db'
if os.path.isfile(sqlite_file):
    os.remove(sqlite_file)

connection = sqlite3.connect(sqlite_file)
sqlite_dao.verify_database(connection, (keeper_schema, thycotic_schema), apply_changes=True)

keeper_record_storage = sqlite.SqliteEntityStorage(lambda: connection, keeper_schema)
keeper_records = []    # type: List[KeeperRecord]
for rec in my_params.record_cache.values():
    record = vault.KeeperRecord.load(my_params, rec)
    if isinstance(record, (vault.PasswordRecord, vault.TypedRecord)):
        kr = KeeperRecord()
        kr.uid = record.record_uid
        kr.title = record.title
        folders = [subfolder.get_folder_path(my_params, x, delimiter='\\') for x in subfolder.find_folders(my_params, kr.uid) if x]
        if len(folders) > 0:
            kr.path = folders[0]
        keeper_records.append(kr)
keeper_record_storage.put(keeper_records)

thycotic_secret_storage = sqlite.SqliteEntityStorage(lambda: connection, thycotic_schema)
thycotic_secrets = []    # type: List[ThycoticSecret]
for secret in thy_secrets:
    ts = ThycoticSecret()
    ts.id = secret['id']
    folder_id = secret['folderId']
    ts.title = secret['name']
    if folder_id in thy_folders:
        ts.path = thy_folders[folder_id]['folderPath']
    thycotic_secrets.append(ts)
thycotic_secret_storage.put(thycotic_secrets)

connection.commit()
connection.close()

logging.info('Done')



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2023 Keeper Security Inc.
# Contact: commander@keepersecurity.com
#
# Example code to retrieve records in a shared folder.
# Usage:
#    python vault.py

import os
import sys
from typing import Optional, Set

from keepercommander import api, record, vault, shared_folder, subfolder
from keepercommander.__main__ import get_params_from_config

# load configuration file.
my_params = get_params_from_config(os.path.join(os.path.dirname(__file__), 'config.json'))

# Prompt for Keeper username if configuration file is empty
if not my_params.user:
    my_params.user = input('User(Email): ')
    if not my_params.user:
        sys.exit(1)

# Login to Keeper
api.login(my_params)
if not my_params.session_token:
    sys.exit(1)

# Load Vault data
api.sync_down(my_params)

# Keeper supports several record formats internally
# Version 2 or Legacy or General records, V2
# Version 3 or Typed records, V3
# Version 4 or File records, V4
# ...
# Record versions 3 and up replace legacy v2 record format

# Commander has 2 class sets for Keeper records
# 1. "Record" class in the record.py
#    This class represents V2 records originally.
#    When used with V3 records it maps V3 record fields to V2 data structure.
#    Convenient when reading records
#    Use api.get_record(param, record_uid) method to load record
# 2. Newer record classes in the vault.py
#    "KeeperRecord" the base class for all record types
#    "PasswordRecord" represents V2 or legacy records
#    "TypedRecord" - V3 or typed records
#    "FileRecord" - V4 or file attachment
#    Use vault.KeeperRecord.load(params, record_uid) to load a record

# Old record class example
# Enumerated all records in all shared folders.
# Shared folders are stored in "shared_folder_cache" attribute of KeeperParams object

old_record: Optional[record.Record] = None
if my_params.shared_folder_cache:
    shared_folder: Optional[shared_folder.SharedFolder]
    for shared_folder_uid in my_params.shared_folder_cache:
        # Load shared folder
        shared_folder = api.get_shared_folder(my_params, shared_folder_uid)
        if shared_folder:
            record_no = len(shared_folder.records) if shared_folder.records else 0
            print(f'Shared Folder "{shared_folder.name}" has {record_no} record(s)')
            for record_key in shared_folder.records:
                record_uid = record_key['record_uid']
                # Load a record into old record class
                old_record = api.get_record(my_params, record_uid)
                if old_record:    # V2 or V3
                    print(f'Record "{old_record.title}" has {("" if old_record.password else "no ")}password')

# New record class example
new_record: Optional[vault.KeeperRecord] = None

# get a random shared folder/subfolder in the vault tree
folder: Optional[subfolder.BaseFolderNode] = None
records: Optional[Set[str]] = None
for f_uid, recs in my_params.subfolder_record_cache.items():
    if not f_uid:    # Root folder does not have Folder UID
        continue
    if not recs:   # folder has no records
        continue
    if f_uid not in my_params.folder_cache:
        continue
    f: Optional[subfolder.BaseFolderNode] = my_params.folder_cache[f_uid]    # folder description
    # pick a shared folder/subfolder
    if f.type in (subfolder.BaseFolderNode.SharedFolderType, subfolder.BaseFolderNode.SharedFolderFolderType):
        folder = f
        records = recs
        break

if folder and records:
    print(f'Folder "{folder.name}" has {len(records)} record(s)')
    for record_uid in records:
        # Load record
        new_record = vault.KeeperRecord.load(my_params, record_uid)
        if isinstance(new_record, vault.PasswordRecord):
            print(f'Legacy Record "{new_record.title}" has {("" if new_record.password else "no ")}password')
        elif isinstance(new_record, vault.TypedRecord):
            password_field = new_record.get_typed_field('password')
            if password_field:
                password = password_field.get_default_value(str)
                print(f'Record "{new_record.title}" type "{new_record.record_type}" has {("" if password else "no ")}password')
            else:
                print(f'Record "{new_record.title}" type "{new_record.record_type}" does not have password field')
        elif new_record:
            print(f'Record "{new_record.title}" type "{new_record.record_type}" does not have password field')
        else:
            print(f'Record UID "{record_uid}" is not supported')







from __future__ import annotations

from typing import Dict, Any
from dataclasses import dataclass
import enum
import hashlib
import hmac

from .. import utils
from ..proto import pedm_pb2

@dataclass
class DeploymentAgentInformation:
    hash_key: bytes
    peer_public_key: bytes

    def to_dict(self) -> Dict[str, Any]:
        return {
            'hash_key': utils.base64_url_encode(self.hash_key),
            'peer_public_key': utils.base64_url_encode(self.peer_public_key),
        }

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> DeploymentAgentInformation:
        hash_key = data['hash_key']
        peer_public_key = data.get('peer_public_key') or data['pair_public_key']
        return DeploymentAgentInformation(hash_key=utils.base64_url_decode(hash_key),
                                          peer_public_key=utils.base64_url_decode(peer_public_key))


def get_collection_uid(hash_key: bytes, collection_type: int, value: str) -> str:
    message = collection_type.to_bytes(4, byteorder='big') + value.strip().lower().encode('utf-8')
    d = hmac.new(hash_key, message, hashlib.sha256).digest()
    x1 = int.from_bytes(d[:16], byteorder='big', signed=False)
    x2 = int.from_bytes(d[16:], byteorder='big', signed=False)
    return utils.base64_url_encode((x1 ^ x2).to_bytes(length=16, byteorder='big', signed=False))


class CollectionType(int, enum.Enum):
    Other = 0,
    OsBuild = 1,
    Application = 2,
    UserAccount = 3,
    GroupAccount = 4,
    ApplicationName = 5,
    UserName = 10,
    CustomAppCollection = 102,
    CustomUserCollection = 103,
    CustomMachineCollection = 201,
    OsVersion = 202,


def collection_type_to_name(collection_type: int) -> str:
    if collection_type == CollectionType.OsBuild:
        return 'OS Build'
    if collection_type == CollectionType.Application:
        return 'Application'
    if collection_type == CollectionType.UserAccount:
        return 'User Account'
    if collection_type == CollectionType.GroupAccount:
        return 'Group Account'
    if collection_type == CollectionType.ApplicationName:
        return 'App Name'
    if collection_type == CollectionType.UserName:
        return 'User Name'
    if collection_type == CollectionType.CustomAppCollection:
        return 'App Collection'
    if collection_type == CollectionType.CustomUserCollection:
        return 'User Collection'
    if collection_type == CollectionType.CustomMachineCollection:
        return 'Machine Collection'
    if collection_type == CollectionType.OsVersion:
        return 'OS Version'
    return 'Other'

def collection_link_type_to_name(collection_link_type: int) -> str:
    if collection_link_type == pedm_pb2.CollectionLinkType.CLT_AGENT:
        return 'AGENT'
    if collection_link_type == pedm_pb2.CollectionLinkType.CLT_POLICY:
        return 'POLICY'
    if collection_link_type == pedm_pb2.CollectionLinkType.CLT_COLLECTION:
        return 'COLLECTION'
    return 'OTHER'


class EventRequestType(int, enum.Enum):
    Other = 0,
    PrivilegeElevation = 1,
    FileAccess = 2,
    CommandLine = 5,
    LeastPrivilege = 6,
    Custom = 99

def approval_type_to_name(event_type: int) -> str:
    if event_type == EventRequestType.PrivilegeElevation:
        return 'PrivilegeElevation'
    if event_type == EventRequestType.FileAccess:
        return 'FileAccess'
    if event_type == EventRequestType.CommandLine:
        return 'CommandLine'
    if event_type == EventRequestType.LeastPrivilege:
        return 'LeastPrivilege'
    if event_type == EventRequestType.Custom:
        return 'Custom'
    return 'Other'



from __future__ import annotations
import logging
from .infrastructure import Infrastructure
from .record_link import RecordLink
from .user_service import UserService
from .constants import PAM_MACHINE, PAM_DIRECTORY
from .utils import get_connection
from .types import DiscoveryObject
from ..keeper_dag import EdgeType
from ..keeper_dag.edge import DAGEdge
import re
import sys
from typing import Any, Optional, Callable, TYPE_CHECKING

if TYPE_CHECKING:
    from ..keeper_dag.vertex import DAGVertex


class Verify:

    """
    Check the graphs to make sure they are correct.

    This class will try to find problems with the graph, and then try to fix them, if flagged to do so.
    Checks are:

    * Check if the user services and task graph matches what Discovery found for the machine.
    * Check if the infrastructure graph has actual record for the record UID stored in the content.

    """

    USER_SERVICE = "User Service/Task Mapping"
    COLOR_RESET = "reset"
    OK = "ok"
    FAIL = "fail"
    UNK = "unk"
    TITLE = "title"

    def __init__(self, record: Any, logger: Optional[Any] = None, debug_level: int = 0,
                 output: Optional = None, colors: Optional[dict] = None, **kwargs):

        if output is None:
            output = sys.stderr
        self.output = output

        if colors is None:
            colors = {}
        self.colors = colors

        self.conn = get_connection(**kwargs)

        self.record = record

        # Load all the Infrastructure graph, starting at sync point 0
        self.infra = Infrastructure(record=record, logger=logger, debug_level=debug_level, fail_on_corrupt=False,
                                    **kwargs)
        self.infra.load(sync_point=0)

        self.record_link = RecordLink(record=record, logger=logger, debug_level=debug_level, fail_on_corrupt=False,
                                      **kwargs)
        self.user_service = UserService(record=record, logger=logger, debug_level=debug_level, fail_on_corrupt=False,
                                        **kwargs)

        if logger is None:
            logger = logging.getLogger()
        self.logger = logger
        self.debug_level = debug_level
        self.logger.debug(f"configuration uid is {self.conn.get_record_uid(record)}")

    def _msg(self, msg, color_name="NONE"):
        print(f"{self.colors.get(color_name, '')}{msg}{self.colors.get(Verify.COLOR_RESET, '')}", file=self.output)

    def run(self, fix: bool = False, lookup_record_func: Optional[Callable] = None):

        self.verify_infra_dag_connections(fix=fix)
        self.verify_user_service(fix=fix)
        if lookup_record_func is not None:
            self.verify_record_exists(fix=fix, lookup_record_func=lookup_record_func)

    @staticmethod
    def _split_user(user: str, hostname: Optional[str] = None, host: Optional[str] = None):
        domain = None
        if "\\" in user:
            domain, user = user.split("\\", 1)
            if domain == ".":
                domain = None
        elif "@" in user:
            user, domain = user.split("@", 1)
        if domain is not None and hostname is not None:
            domain = domain.lower()

            # Don't use IP addresses
            if re.match(r'\d+\.\d+.\d+\.\d+', host) is not None:
                host = None

            if hostname is not None:
                hostname = hostname.lower()
                if domain == hostname:
                    domain = None
                elif domain == hostname.split(".")[0]:
                    domain = None

            if host is not None:
                host = host.lower()
                if domain == host:
                    domain = None
                elif domain == host.split(".")[0]:
                    domain = None

        return user, domain

    def _find_infra_user_vertex(self, resource_vertex: DAGVertex, user: str, domain: Optional[str] = None) -> (
            Optional)[DAGVertex]:

        user = user.lower()
        resource_content = DiscoveryObject.get_discovery_object(resource_vertex)

        # If the domain is None, assume it a local user.
        if domain is None:
            self.logger.debug("    no domain, assume local user")
            for user_vertex in resource_vertex.has_vertices():
                content = DiscoveryObject.get_discovery_object(user_vertex)
                self.logger.debug(f"    * {content.name}, {content.item.user}")
                if content.name.lower() == user:
                    self.logger.debug("        MATCH")
                    return user_vertex
                hostname = None
                if resource_content.record_type == PAM_MACHINE:
                    hostname = resource_content.item.facts.name
                child_user, child_domain = self._split_user(
                    user=content.item.user,
                    hostname=hostname,
                    host=resource_content.item.host)
                if user == child_user and child_domain is None:
                    self.logger.debug("        MATCH")
                    return user_vertex
            return None

        self.logger.debug("    has domain, assume directory user")

        configuration_vertex = self.infra.get_configuration
        for vertex in configuration_vertex.has_vertices():
            content = DiscoveryObject.get_discovery_object(vertex)
            if content.record_type != PAM_DIRECTORY:
                continue
            if content.name.lower() == domain.lower():
                for user_vertex in vertex.has_vertices():
                    user_content = DiscoveryObject.get_discovery_object(user_vertex)
                    if user_content.name.lower() == user or user_content.item.user.lower() == user:
                        return user_vertex

        return None

    def _fix_user_service_acl(self, resource_content: DiscoveryObject, user_vertex: DAGVertex, acl_type: str,
                              fix: bool = False) -> bool:

        user_content = DiscoveryObject.get_discovery_object(user_vertex)
        user_record_uid = user_content.record_uid
        if user_record_uid is not None:
            acl = self.user_service.get_acl(resource_content.record_uid, user_record_uid)
            if acl is not None:
                flag = getattr(acl, acl_type)
                if flag is False:

                    self._msg(f"   - user {user_content.name}, {user_record_uid} is "
                              f"missing an ACL type {acl_type} to "
                              f"machine {resource_content.name}")
                    if fix is True:
                        self._msg(f"     added {acl_type} to the ACL between "
                                  f"user {user_content.name}, {user_record_uid} and "
                                  f"machine {resource_content.name}", color_name=Verify.OK)
                        setattr(acl, acl_type, True)
                        self.user_service.belongs_to(resource_content.record_uid, user_record_uid, acl=acl)
                        return True
                    else:
                        self._msg(f"     not fixing user", color_name=Verify.FAIL)
                else:
                    self.logger.debug(f"user service ACL does have is_service as True")
            else:
                self.logger.debug(f"there is no ACL between the user and the resource")
        else:
            self.logger.debug(f"use does not have a record yet")

        return False

    def _get_infra_configuration(self):

        # Check to make sure the user service graph exists.
        # The "UserService" instance should do this, but we want to make sure.
        if self.user_service.dag.has_graph is False:
            self.logger.debug("the user service graph contains no data")
            configuration_vertex = self.user_service.dag.get_root
            if configuration_vertex.uid != self.conn.get_record_uid(self.record):
                raise Exception("The user service graph root/con does not match ")

        return self.infra.get_configuration

    def verify_user_service(self, fix: bool = False):

        """

        """

        # STEP 1 - Make sure UserService graph matches Infrastructure

        self._msg("\nChecking if user service/task graph matches infrastructure.\n\n", color_name="title")

        were_fixes = False

        infra_configuration = self._get_infra_configuration()
        for resource_vertex in infra_configuration.has_vertices():

            resource_content = DiscoveryObject.get_discovery_object(resource_vertex)

            if resource_content.record_type != PAM_MACHINE or resource_content.record_uid is None:
                self._msg(f" * Machine {resource_content.name} does not have record UID, yet.",
                          color_name=Verify.UNK)
                continue

            user_service_resource_vertex = self.user_service.dag.get_vertex(resource_content.record_uid)
            if user_service_resource_vertex is None:
                self._msg(f" * Machine {resource_content.name} does not have a vertex in the user service graph")
                if fix is True:
                    user_service_resource_vertex = self.user_service.dag.add_vertex(resource_content.record_uid)

                    record_uid = self.conn.get_record_uid(self.record)
                    self.user_service.belongs_to(record_uid, resource_content.record_uid, acl=None)

                    self._msg(f"   added vertex for machine {resource_content.name}, and linked "
                              " to configuration.", color_name=Verify.OK)
                else:
                    self._msg(f"   not fixing, skip this resource.", color_name=Verify.FAIL)
                    continue

            if self.user_service.resource_has_link(resource_content.record_uid) is False:
                self._msg(f" * Machine {resource_content.name} is not linked to the configuration.")

                if fix is True:
                    user_service_resource_vertex.belongs_to(self.user_service.dag.get_root, edge_type=EdgeType.LINK)
                    self._msg(f"   linking machine {resource_content.name} to the configuration",
                              color_name=Verify.OK)
                else:
                    self._msg(f"   not fixing, skip this resource.", color_name=Verify.FAIL)
                    continue

            self.logger.debug(f"found machine: {resource_content.name}, {resource_content.record_uid}")

            for item in resource_content.item.facts.services:
                user, domain = self._split_user(item.user,
                                                hostname=resource_content.item.facts.name,
                                                host=resource_content.item.host)
                self.logger.debug(f"found service: {item.name}, {user}, {domain}")
                user_vertex = self._find_infra_user_vertex(resource_vertex, user, domain)
                if user_vertex is not None:
                    if self._fix_user_service_acl(resource_content, user_vertex, "is_service", fix=fix) is True:
                        were_fixes = True
                else:
                    self.logger.debug(f"could not find user for the service on the machine")

            for item in resource_content.item.facts.tasks:
                user, domain = self._split_user(item.user,
                                                hostname=resource_content.item.facts.name,
                                                host=resource_content.item.host)
                self.logger.debug(f"  found task: {item.name}, {user}, {domain}")
                user_vertex = self._find_infra_user_vertex(resource_vertex, user, domain)
                if user_vertex is not None:
                    if self._fix_user_service_acl(resource_content, user_vertex, "is_task", fix=fix) is True:
                        were_fixes = True
                else:
                    self.logger.debug(f"    could not find user for the service on the machine")

        if were_fixes is True:
            self._msg("\nSaving fixed user service/task graph.\n\n", color_name=Verify.OK)
            self.user_service.save()

    def verify_record_exists(self, lookup_record_func: Callable, fix: bool = False):
        """
        This will verify that a record exists for infrastructure content.
        """

        self._msg("\nChecking if infrastructure records exist.\n\n", color_name=Verify.TITLE)

        infra_configuration = self._get_infra_configuration()

        def _check(vertex: DAGVertex, indent: int = 0) -> bool:

            fixes = False
            pad = "".ljust(4 * indent, ' ')

            if not vertex.active:
                return False
            if vertex.has_data is True:
                content = DiscoveryObject.get_discovery_object(vertex)

                self._msg(f"{pad}* {content.record_uid or 'NA'}, {content.title}", color_name="title")

                if content.record_uid is not None:
                    record = lookup_record_func(content.record_uid)
                    if record is None:
                        self._msg(f"{pad}  did not have a record.", color_name="fail")
                        if fix is True:
                            content.record_uid = None
                            vertex.add_data(content)
                            self._msg(f"{pad}  remove record uid from graph.", color_name=Verify.OK)
                            fixes = True
                else:
                    self._msg(f"{pad}  has no record uid.", color_name="unk")

            for next_vertex in vertex.has_vertices():
                if next_vertex.uid == vertex.uid:
                    self._msg(f"{pad}  * this vertex loops to itself!", color_name=Verify.FAIL)
                    continue
                edge_type = next_vertex.get_highest_edge_version(vertex.uid)
                if edge_type != EdgeType.DELETION:
                    if _check(next_vertex,  indent=indent + 1) is True:
                        fixes = True

            return fixes

        were_fixes = _check(infra_configuration, indent=0)

        if were_fixes is True:
            self._msg("\nSaving fixed record uids in infrastructure graph.\n\n", color_name=Verify.OK)
            self.infra.save(delta_graph=False)

    def verify_infra_dag_connections(self, fix: bool = False):

        self._msg("\nChecking if infrastructure vertex refs loop.\n", color_name=Verify.TITLE)

        def _check(vertex: DAGVertex, indent: int = 0):
            pad = "".ljust(4 * indent, ' ')

            fixes = False

            text = ""
            if vertex.active is False:
                text += f"(Inactive) "
            elif vertex.corrupt is True:
                text += f"{self.colors.get(Verify.FAIL)}Corrupt{self.colors.get(Verify.COLOR_RESET, '')}"
            elif vertex.has_data is True:
                content = DiscoveryObject.get_discovery_object(vertex)
                text += content.title

            self._msg(f"{pad}checking {vertex.uid}; {text}")
            for edge in vertex.edges:  # type: DAGEdge
                if edge.edge_type == EdgeType.DATA:
                    self._msg(f"{pad}  * found data edge")
                else:
                    self._msg(f"{pad}  * edge {edge.edge_type} to {edge.head_uid} (parent/belongs_to)")
                    if edge.head_uid == vertex.uid:
                        if edge.edge_type == EdgeType.DELETION:
                            self._msg(f"{pad}  * found DELETION of DATA edge")
                            continue
                        self._msg(f"{pad}  * vertex as a non-DATA edge looping to self", color_name=Verify.FAIL)
                        if fix is True:
                            self._msg(f"{pad}  * deleting key", color_name=Verify.OK)
                            fixes = True
                            edge.delete()

            # Get all the child vertices, allow self ref, so we can delete it if not already deleted.
            for next_vertex in vertex.has_vertices(allow_self_ref=True):
                if next_vertex.uid == vertex.uid:
                    version, edge = next_vertex.get_highest_edge_version(vertex.uid)
                    if edge.edge_type == EdgeType.DELETION:
                        continue
                    else:
                        self._msg(f"{pad}    * next vertex references itself", color_name=Verify.FAIL)
                        if fix is True:
                            self._msg(f"{pad}    * delete this reference", color_name=Verify.OK)
                            fixes = True
                            edge.delete()
                        else:
                            self._msg(f"{pad}  * not fixing, however skipping to prevent loop",
                                      color_name=Verify.FAIL)
                            continue

                self._msg(f"{pad}  next vertex is {next_vertex.uid}")
                if _check(next_vertex, indent + 1) is True:
                    fixes = True

            return fixes

        configuration = self.infra.get_configuration
        were_fixes = _check(configuration, 0)

        if were_fixes is True:
            self._msg("\nSaving graph vertices references.\n\n", color_name=Verify.OK)
            self.infra.save(delta_graph=False)



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2024 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

import argparse
from ...params import KeeperParams
from ...commands.base import report_output_parser, Command
from ..core.service_manager import ServiceManager
from ..core.globals import init_globals
from ..decorators.logging import debug_decorator

class StartService(Command):
    """Command to start the service."""
    @debug_decorator
    def get_parser(self):
        parser = argparse.ArgumentParser(prog='service-start', parents=[report_output_parser], description='Starts the Commander API service with existing configuration')
        return parser

    def execute(self, params: KeeperParams, **kwargs) -> None:
        init_globals(params)
        ServiceManager.start_service()

class StopService(Command):
    """Command to stop the service."""
    @debug_decorator
    def get_parser(self):
        parser = argparse.ArgumentParser(prog='service-stop', parents=[report_output_parser], description='Stops the Commander API service currently running')
        return parser
    
    def execute(self, params: KeeperParams, **kwargs) -> None:
        ServiceManager.stop_service()

class ServiceStatus(Command):
    """Command to get service status."""
    @debug_decorator
    def get_parser(self):
        parser = argparse.ArgumentParser(prog='service-status', parents=[report_output_parser], description='Displays if the Commander API service is running or stopped')
        return parser
    
    def execute(self, params: KeeperParams, **kwargs) -> str:
        status = ServiceManager.get_status()
        print(f"Current status: {status}")


# -*- coding: utf-8 -*-
#  _  __  
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|            
#
# Keeper Commander 
# Copyright 2022 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

import psycopg2
import logging

"""Commander Plugin for Postgres Database Server
   Dependencies: 
       pip3 install psycopg2-binary
"""


class Rotator:
    def __init__(self, host, login, password, port=5432, db='postgres', **kwargs):
        self.host = host
        self.login = login
        self.password = password
        self.port = port
        self.db = db

    def rotate_start_msg(self):
        """Display msg before starting rotation"""
        logging.info(
            f'Rotating with PostgreSQL plugin on host "{self.host}" and port "{self.port}" using login "{self.login}"'
            f' to connect to db "{self.db}"...'
        )

    def revert(self, record, new_password):
        """Revert rotation of a PostgreSQL password"""
        self.rotate(record, new_password, revert=True)

    def rotate(self, record, new_password, revert=False):
        """Rotate a PostgreSQL password"""
        if revert:
            old_password = new_password
            new_password = self.password
        else:
            old_password = self.password

        try:
            with psycopg2.connect(host=self.host, port=self.port, user=self.login, password=old_password,
                                  database=self.db) as connection:
                logging.debug(f'Connected to {self.host}')
                with connection.cursor() as cursor:
                    sql = f'alter user {self.login} with password %s'
                    cursor.execute(sql, (new_password,))
                    return True
        except Exception as e:
            if revert:
                logging.error('Error reverting password rotation of Postgres server: %s', e)
            else:
                logging.error('Error rotating password of Postgres server: %s', e)

        return False



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2024 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

from typing import Optional, Tuple, Dict, Any
from flask import request, jsonify
from html import escape
import tempfile
import os
import json
from ..decorators.logging import logger


class RequestValidator:
    """Shared validation utilities for API requests."""
    
    @staticmethod
    def validate_and_escape_command(request_data: Dict[str, Any]) -> Tuple[Optional[str], Optional[Tuple]]:
        """Validate and escape command from request data.
        
        Args:
            request_data: The request JSON data
            
        Returns:
            Tuple of (escaped_command, error_response) where error_response is None on success
        """
        if not request_data:
            logger.info("Request validation failed: No JSON data provided")
            return None, (jsonify({"status": "error", "error": "No JSON data provided"}), 400)
        
        command = request_data.get("command")
        if not command:
            logger.info("Request validation failed: Missing required field 'command' or incorrect field name")
            return None, (jsonify({"status": "error", "error": "Missing required field \"command\" or incorrect field name"}), 400)
        
        if not isinstance(command, str):
            logger.warning("Request validation failed: Command must be a string")
            return None, (jsonify({"status": "error", "error": "Command must be a string"}), 400)
            
        # Escape HTML to prevent XSS
        escaped_command = escape(command)
        logger.debug(f"Command validated and escaped: {escaped_command}")
        return escaped_command, None
    
    @staticmethod
    def process_file_data(request_data: Dict[str, Any], command: str) -> Tuple[str, list]:
        """Process filedata in request and create temporary files, substituting FILEDATA placeholders.
        
        Args:
            request_data: The request JSON data
            command: The command string to process
            
        Returns:
            Tuple of (processed_command, temp_file_paths)
        """
        import re
        temp_files = []
        processed_command = command
        
        # Check if command contains FILEDATA placeholder and request has filedata
        if "FILEDATA" in command and "filedata" in request_data:
            filedata = request_data.get("filedata")
            if not isinstance(filedata, dict):
                logger.warning("filedata must be a JSON object")
                return processed_command, temp_files
            
            try:
                # Create temporary file with the filedata content
                with tempfile.NamedTemporaryFile(mode='w', suffix='.json', delete=False, encoding='utf-8') as temp_file:
                    json.dump(filedata, temp_file, indent=2)
                    temp_file_path = temp_file.name
                    temp_files.append(temp_file_path)
                    
                    # Replace FILEDATA placeholder with temporary file path, but only when it's used as a file parameter
                    # Supports: PAM import (--filename), Import (positional), Enterprise-push (positional)
                    file_param_pattern = r'(--filename[=\s]+["\']?)FILEDATA(["\']?)'
                    processed_command = re.sub(file_param_pattern, f'\\1{temp_file_path}\\2', processed_command)
                    
                    # Also handle standalone FILEDATA (not in quotes)
                    standalone_pattern = r'\bFILEDATA\b(?!["\'])'
                    processed_command = re.sub(standalone_pattern, temp_file_path, processed_command)
                    
                    logger.debug(f"Created temporary file {temp_file_path} for filedata")
                    
            except Exception as e:
                logger.error(f"Error creating temporary file for filedata: {e}")
                # Clean up any created temp files
                for temp_path in temp_files:
                    try:
                        os.unlink(temp_path)
                    except Exception:
                        pass
                return command, []
        
        return processed_command, temp_files
    
    @staticmethod
    def cleanup_temp_files(temp_files: list) -> None:
        """Clean up temporary files.
        
        Args:
            temp_files: List of temporary file paths to clean up
        """
        for temp_path in temp_files:
            try:
                if os.path.exists(temp_path):
                    os.unlink(temp_path)
                    logger.debug(f"Cleaned up temporary file: {temp_path}")
            except Exception as e:
                logger.warning(f"Failed to clean up temporary file {temp_path}: {e}")
    
    @staticmethod
    def validate_request_json() -> Optional[Tuple]:
        """Validate that request contains valid JSON.
        
        Returns:
            Error response tuple if validation fails, None if valid
        """
        if not request.is_json:
            logger.info("Request validation failed: Content-Type must be application/json")
            return jsonify({"status": "error", "error": "Content-Type must be application/json"}), 400
        
        if not request.json:
            logger.info("Request validation failed: Invalid or empty JSON")
            return jsonify({"status": "error", "error": "Invalid or empty JSON"}), 400
        
        return None



import json
import logging
import urllib
from typing import Union, List, Dict, Optional
from urllib import parse

from . import utils, crypto
from .params import KeeperParams
from .proto import APIRequest_pb2, client_pb2, record_pb2
from .utils import is_pw_strong
from .vault import KeeperRecord, TypedRecord

def has_passkey(record):   # type: (KeeperRecord) -> bool
    if not isinstance(record, TypedRecord) or not record.get_typed_field('passkey'):
        return False
    return bool(record.get_typed_field('passkey').value)

def _get_pass(record):  # type: (KeeperRecord) -> Union[str, None]
    from .breachwatch import BreachWatch
    return BreachWatch.extract_password(record) or None

def get_security_score(record): # type: (KeeperRecord) -> Union[int, None]
    password = _get_pass(record)
    return None if not password \
        else utils.password_score(password) or has_passkey(record) and 100 or 0

def encrypt_security_data(params, data):
    if params.forbid_rsa and not params.enterprise_ec_key:
        raise Exception('Enterprise ECC public key is not available')

    if not params.forbid_rsa and not params.enterprise_rsa_key:
        raise Exception('Enterprise RSA public key is not available')

    data = json.dumps(data).encode('utf8')
    pubkey = params.enterprise_ec_key if params.forbid_rsa else params.enterprise_rsa_key
    encrypt_fn = crypto.encrypt_ec if params.forbid_rsa else crypto.encrypt_rsa
    return encrypt_fn(data, pubkey)

def prep_security_data(params, record):
    get_bw_obj = lambda rec: next(
        iter(params.breach_watch_records.get(rec, {}).get('data_unencrypted', {}).get('passwords', [])),
        {}) if params.breach_watch else None

    from .breachwatch import BreachWatch
    score = get_security_score(record)
    # Send empty object to remove old security data (when password and/or passkey are removed)
    sec_data = b''
    if score is not None:
        sec_data = {'strength': score}
        password = _get_pass(record)
        login_url = BreachWatch.extract_url(record)
        parse_results = urllib.parse.urlparse(login_url)
        domain = parse_results.hostname or parse_results.path
        pw_obj = get_bw_obj(record.record_uid)
        if pw_obj is not None and password:
            status = pw_obj.get('status')
            sec_data['bw_result'] = client_pb2.BWStatus.Value(status) if status \
                else client_pb2.BWStatus.GOOD if is_pw_strong(score) \
                else client_pb2.BWStatus.WEAK
        if domain:
            sec_data.update(dict(domain=domain))
            data_size = len(json.dumps(sec_data).encode('utf8'))
            max_size = 244
            diff = max_size - data_size
            # truncate domain string if needed to avoid reaching RSA encryption data size limitation
            if diff < 0:
                new_length = len(domain) + diff
                sec_data.update(dict(domain=domain[:new_length]))
        sec_data = encrypt_security_data(params, sec_data)
    return sec_data

def prep_security_data_update(params, record): # type: (KeeperParams, KeeperRecord) -> Optional[APIRequest_pb2.SecurityData]
    sd = APIRequest_pb2.SecurityData()
    try:
        sd.uid = utils.base64_url_decode(record.record_uid)
        data = prep_security_data(params, record)
    except:
        logging.error('Could not update security data for record')
        return
    if data:
        sd.data = data
    return sd

def prep_score_data(record):
    empty_score_data = crypto.encrypt_aes_v2(json.dumps(dict()).encode('utf8'), record.record_key)
    score = get_security_score(record)
    if score is None:
        return empty_score_data

    try:
        password = _get_pass(record)
        pad_length = max(25 - len(password), 0) if password else 0
        pad = ''.join([' ' for _ in range(pad_length)])
        score_data = dict(version=1, password=password, score=score, padding=pad)
        data = json.dumps(score_data).encode('utf-8')
        sec_score_data = crypto.encrypt_aes_v2(data, record.record_key)
    except:
        logging.error(f'Could not calculate security score data for record')
        sec_score_data = empty_score_data
    return sec_score_data

def prep_score_data_update(params, record):    # type: (KeeperParams, KeeperRecord) -> APIRequest_pb2.SecurityScoreData
    ssd = APIRequest_pb2.SecurityScoreData()
    revision = params.security_score_data.get(record.record_uid, {}).get('revision')
    ssd.uid = utils.base64_url_decode(record.record_uid)
    ssd.data = prep_score_data(record)
    if revision:
        ssd.revision = revision
    return ssd

def needs_security_audit(params, record):  # type: (KeeperParams, KeeperRecord) -> bool
    if not params.enterprise_ec_key or not record:
        return False

    saved_score_data = params.security_score_data.get(record.record_uid, {})
    saved_sec_data = params.breach_watch_security_data.get(record.record_uid, {})
    score_data =  saved_score_data.get('data', {})
    current_password = _get_pass(record)
    if current_password != score_data.get('password') or None:
        return True

    scores = dict(new=get_security_score(record) or 0, old=score_data.get('score', 0))
    score_changed_on_passkey = any(x >= 100 for x in scores.values()) and any(x < 100 for x in scores.values())
    creds_removed = bool(scores.get('old') and not scores.get('new'))
    needs_alignment = bool(scores.get('new')) and not saved_sec_data
    return score_changed_on_passkey or creds_removed or needs_alignment

def update_security_audit_data(params, records):   # type: (KeeperParams, List[KeeperRecord]) -> int
    if not params.enterprise_ec_key:
        return 0

    from . import api
    update_limit = 1000
    total_updates = len(records)
    failed_updates = []
    while records:
        chunk = records[:update_limit]
        records = records[update_limit:]
        rq = APIRequest_pb2.SecurityDataRequest()
        rq.encryptionType = get_security_data_key_type(params)
        try:
            sec_data_objs = (prep_security_data_update(params, rec) for rec in chunk)
            score_data_objs = (prep_score_data_update(params, rec) for rec in chunk)
            rq.recordSecurityData.extend(sd for sd in sec_data_objs if sd)
            rq.recordSecurityScoreData.extend(sd for sd in score_data_objs if sd)
            rs = api.communicate_rest(params, rq, 'enterprise/update_security_data')
        except:
            failed_updates.extend(chunk)

    if failed_updates:
        logging.error(f'Could not update security data for {len(failed_updates)} records')

    return total_updates - len(failed_updates)

def attach_security_data(params, record, rq_param):
    # type: (KeeperParams, Union[str, Dict[str, any], KeeperRecord], Union[record_pb2.RecordUpdate, record_pb2.RecordAdd]) -> Union[record_pb2.RecordUpdate, record_pb2.RecordAdd]
    try:
        if not isinstance(record, TypedRecord):
            if isinstance(record, dict):
                record['version'] = record.get('version', 3)
            record = KeeperRecord.load(params, record)
        if needs_security_audit(params, record):
            rq_param.securityData.data = prep_security_data(params, record)
            rq_param.securityScoreData.data = prep_score_data(record)
    except:
        pass
    finally:
        return rq_param

def get_security_data_key_type(params):
    return record_pb2.ENCRYPTED_BY_PUBLIC_KEY_ECC if params.forbid_rsa \
        else record_pb2.ENCRYPTED_BY_PUBLIC_KEY



#!/usr/bin/env python3
#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2021 Keeper Security Inc.
# Contact: commander@keepersecurity.com
#
# Example showing how to add a new or existing record 
# to an existing shared folder. 
#

import getopt
import os
import sys

from keepercommander import api
from keepercommander.__main__ import get_params_from_config
from keepercommander.commands.recordv2 import RecordAddCommand, RecordEditCommand
from keepercommander.params import KeeperParams


def login_to_keeper_with_config(filename):   # type: (str) -> KeeperParams
    # Load connection parameters from file
    my_params = get_params_from_config(filename)
    my_params.password = ''  # Keeper account password. Can be omitted if account is setup for session resumption
    # Login to Keeper
    api.login(my_params)
    api.sync_down(my_params)
    return my_params


def create_or_update_record(params, criteria):
    # Scan vault for shared folder

    sfs = api.search_shared_folders(params, criteria)
    if len(sfs) == 0:
        print(f'Shared folder {criteria} not found')
        return
    shared_folder = sfs[0]
    record_title = 'Test Record'
    record_uid = None
    if shared_folder.records:
        records = [api.get_record(params, x['record_uid']) for x in shared_folder.records]
        record_uid = next((x.record_uid for x in records if x.title.casefold() == record_title.casefold()), None)

    shared_folder_uid = shared_folder.shared_folder_uid
    if record_uid:
        command = RecordEditCommand()
        command.execute(params, record=record_uid, generate=True)
        print(f'Updated record {record_uid} in shared folder uid={shared_folder_uid}')
    else:
        command = RecordAddCommand()
        # Inputs - hard coded for demo purposes
        record_uid = command.execute(
            params,
            title='Test Record',
            login='someone@company.com',
            folder=shared_folder_uid,
            generate=True,
            # force=True disables asking for missing fields
            force=True)
        print(f'Added record {record_uid} to shared folder uid={shared_folder_uid}')


if __name__ == '__main__':
    opts, args = getopt.getopt(sys.argv[1:], 'c:', ['config='])
    if len(args) == 0:
        print('Shared folder name or UID parameter is required.')
        sys.exit(1)

    config_file = 'myconfig.json'
    for p_name, p_value in opts:
        if p_name == '-c' or p_name == '--config':
            config_file = p_value

    if not os.path.exists(config_file):
        print(f'Config file {config_file} not found')
        sys.exit(1)

    # Login and sync
    params = login_to_keeper_with_config(config_file)

    # Add record to shared folder
    create_or_update_record(params, args[0])
